"""Парсинг прошедших работ."""
from time import time

from pyaklass.parse.func import fetch_and_parse


def parse_completed_works(testworks: list, elements: list):
    """Спарсить информацию о прошедших работах.

    Args:
        testworks (list): Список, в который добавляются работы
        elements (list): Список `<tr>`-элементов
    """
    for elem in elements:
        link_elem = elem.css_first('a')

        href = link_elem.attrs['href']
        test_id = int(href[href.rindex('/') + 1:href.rfind('?')])

        submit_unix = int(elem.css_first('[data-utc-date]').attrs['data-utc-date'][:-3])
        subject = elem.css_first('.subject').text(deep=False, strip=True)

        testworks.append({
            'id': test_id,
            'title': link_elem.text(deep=False, strip=True),
            'subject': subject,
            'submit_unix': submit_unix,
        })


def parse_completed_work_details(testwork: dict, testwork_page) -> bool:
    """Спарсить дополнительные сведения о прошедшей работе.

    Args:
        testwork (dict): Работа
        testwork_page (LexborHTMLParser): HTML-страница со сведениями

    Returns:
        bool: Доступна ли работа для сдачи в данный момент?
    """
    info_el = testwork_page.css_first('#testwork')
    if not info_el:
        return False

    script_text: str = testwork_page.css_first('body > script:last-of-type').text(deep=False)

    deadline_unix = script_text[script_text.index('endUTC:') + 7:]
    deadline_unix = float(deadline_unix[:deadline_unix.index(',')]) / 1000

    max_attempts = script_text[script_text.index('triesAllowed:"') + 14:]
    max_attempts = int(max_attempts[:max_attempts.index('"')])

    results_el = testwork_page.css_first('.data-row')
    if results_el.css_first('.data-content > em'):  # The work has been skipped
        results_id = None
        exercises = int(testwork_page.css('.adjust')[-1].text(deep=False))
    else:
        results_id = int(results_el.attrs['data-test-result-id'])
        exercises = int(testwork_page.css('[data-ex-pos]')[-1].attrs['data-ex-pos'])

    testwork.update({
        'deadline_unix': deadline_unix,
        'max_attempts': max_attempts,
        'results_id': results_id,
        'exercises': exercises,
    })
    return (time() - deadline_unix) >= 1


async def fetch_and_parse_completed_work_details(client, limiter, testwork: dict):
    """Получить и парсить дополнительные сведения о прошедшей работе.

    Args:
        client (AsyncClient): Клиент
        limiter (tuple): Ограничитель и его синхронизатор
        testwork (dict): Работа

    Returns:
        Union[dict, None]: Работа или None в случае, если работа недоступна для сдачи
    """
    testwork_page = await fetch_and_parse(
        client, limiter, f"TestWork/Results/{testwork['id']}?from=%2Ftestwork",
    )
    if testwork_page and parse_completed_work_details(testwork, testwork_page):
        return testwork



"""Исключения для PYaKlass."""


class PyaklassError(Exception):
    """Базовый класс для исключений."""

    def __init__(self, reason: str):
        """Инициализировать исключение.

        Args:
            reason (str): Причина
        """
        self.message = reason

    def __str__(self) -> str:
        """Вернуть удобно-читаемое представление исключения.

        Returns:
            str: Сообщение исключения
        """
        return self.message


class LoginError(PyaklassError):
    """Ошибка входа."""


class PossibleIPBlockError(PyaklassError):
    """Ошибка из-за возможной блокировки IP сервисом."""

    def __init__(self):
        """Инициализировать исключение."""
        super().__init__('Инициализировано исключение для предотвращения блокировки IP сервисом.')



"""Внутренние функции для получения информации о работах."""
from asyncio import gather
from time import time

from pyaklass import parse


async def get_testworks(client, limiter: tuple, testworks_page=None, remove_old=False) -> list:
    """Получить работы.

    Args:
        client (httpx.AsyncClient): Авторизированный клиент
        limiter (tuple): Ограничитель и его синхронизатор
        testworks_page (LexbotHTMLParser): HTML-страница с работами
        remove_old (bool): Игнорировать работы, заданные больше 30 дней назад?

    Returns:
        list: Работы
    """
    testworks = []

    if not testworks_page:
        testworks_page = await parse.func.fetch_and_parse(
            client, limiter, '/testwork?from=menu',
        )
    new_works_el, completed_works_el = testworks_page.css('.wg-testworks')

    if not new_works_el.css_first('.wg-empty'):
        parse.new.parse_new_works(testworks, new_works_el.css('table'))
        testworks = await gather(*(
            parse.new.fetch_and_parse_new_work_details(
                client, limiter, tw,
            ) for tw in testworks
        ))

    if not completed_works_el.css_first('.wg-empty'):
        completed_testworks = []
        parse.completed.parse_completed_works(
            completed_testworks, completed_works_el.css('tbody > tr'),
        )

        if remove_old:
            completed_testworks = tuple(filter(
                lambda tw: (time() - tw.pop('submit_unix')) <= 2592000,
                completed_testworks,
            ))

        completed_testworks = await gather(*(
            parse.completed.fetch_and_parse_completed_work_details(
                client, limiter, tw,
            ) for tw in completed_testworks
        ))
        testworks.extend(filter(lambda tw: tw, completed_testworks))

    return testworks


async def get_results(
    client, limiter: tuple, testwork_id: int, result_id: int, pages: list = None,
) -> list:
    """Получить результаты прошедшей работы.

    Args:
        client (httpx.AsyncClient): Авторизированный клиент
        limiter (tuple): Ограничитель и его синхронизатор
        testwork_id (int): ID прошедшей работы
        result_id (int): ID результатов для работы
        pages (list): HTML-страницы для парсинга

    Returns:
        list: Результаты
    """
    results = []

    if not pages:
        exercises_list_page = await parse.func.fetch_and_parse(
            client,
            limiter,
            f'TestWork/Results/{testwork_id}?from=%2FTestWork',
        )
        exercises_pos = int(
            exercises_list_page.css('.adjust')[-1].text(deep=False, strip=True),
        )

        pages = await gather(*(
            parse.func.fetch_and_parse(
                client,
                limiter,
                (
                    f'TestWork/ExerciseResult?testResultId={result_id}'
                    f'&exercisePosition={pos}&twId={testwork_id}'
                ),
            ) for pos in range(1, exercises_pos + 1)
        ))

    for exercise_page in pages:
        parse.results.parse_work_results(results, exercise_page)

    return results



"""Парсинг активных работ."""
from pyaklass.const import BASE_URL
from pyaklass.parse.func import fetch_and_parse


def parse_new_works(testworks: list, elements: list):
    """Спарсить информацию о новых работах.

    Args:
        testworks (list): Список, в который добавляются работы
        elements (list): Список `<table>`-элементов
    """
    for elem in elements:
        link_elem = elem.css_first('a')

        href = link_elem.attrs['href']
        test_id = int(href[href.rindex('/') + 1:])

        deadline_unix = int(elem.css_first('[data-utc-date]').attrs['data-utc-date'][:-3])
        subject = elem.css_first('.subject').text(deep=False, strip=True)

        testworks.append({
            'id': test_id,
            'title': link_elem.text(deep=False),
            'subject': subject,
            'deadline_unix': deadline_unix,
        })


def parse_new_work_details(testwork: dict, testwork_page):
    """Спарсить дополнительные сведения о новой работе.

    Args:
        testwork (dict): Работа
        testwork_page (LexborHTMLParser): HTML-страница со сведениями
    """
    blockbody_el = testwork_page.css_first('.blockbody')

    max_attempts = int(blockbody_el.css_first('p:last-of-type > .value').text(deep=False))
    themes = [
        (
            f"{BASE_URL}{theme_el.attrs['href']}",
            theme_el.text(deep=False),
        ) for theme_el in blockbody_el.css('.test-work-used-topics a')
    ]

    testwork.update({
        'max_attempts': max_attempts,
        'themes': themes,
    })


async def fetch_and_parse_new_work_details(client, limiter: tuple, testwork: dict):
    """Получить и парсить дополнительные сведения о новой работе.

    Args:
        client (AsyncClient): Клиент
        limiter (tuple): Ограничитель и его синхронизатор
        testwork (dict): Работа

    Returns:
        dict: Работа
    """
    testwork_page = await fetch_and_parse(
        client, limiter, f"TestWorkRun/Preview/{testwork['id']}",
    )
    parse_new_work_details(testwork, testwork_page)
    return testwork



"""Базовые функции для получения страниц и парсинга."""
from asyncio import sleep
from hashlib import sha256
from time import time

from selectolax.lexbor import LexborHTMLParser

from pyaklass.const import BASE_URL
from pyaklass.exc import PossibleIPBlockError, PyaklassError


def get_captcha_answer(salt: str, target_hash: str) -> int:
    """Получить решение капчи.

    Args:
        salt (str): Соль
        target_hash (str): SHA256 ответа

    Returns:
        int: Решение или 0 в случае провала
    """
    for i in range(0, int(2e5) + 1):
        computed_hash = sha256(f'{i}{salt}{i}'.encode('utf-8'))
        if computed_hash.hexdigest() == target_hash:
            return i

    return 0


async def check_limiter(lock, limiter):
    """Проверить ограничитель.

    Args:
        lock (Lock): Синхронизатор для ограничителя
        limiter (dict): Ограничитель
    """
    with lock:
        value = limiter.value
        if value['current'] >= value['amount']:
            await sleep(value['seconds'])
            value['current'] = 1
        elif time() - (value.get('last_check', 0) + value['seconds'] + 1) >= 1:
            value['current'] = 1
        else:
            value['current'] += 1
        value['last_check'] = time()
        limiter.set(value)


async def solve_captcha(client, limiter: tuple, params) -> str:
    """Решить капчу.

    Args:
        client (httpx.AsyncClient): Клиент
        limiter (tuple): Ограничитель и его синхронизатор
        params (httpx.QueryParams): GET-параметры для решения капчи

    Returns:
        str: Текст HTML-страницы, указанной в `params['returnUrl']`

    Raises:
        PyaklassError: Невозможно решить капчу
    """
    answer = get_captcha_answer(params['base'], params['target'])
    if answer == 0:
        raise PyaklassError(f'Невозможно решить капчу: "{params}".')

    payload = {
        'r': params['returnUrl'],
        'b': params['base'],
        's': answer,
        'tn': 0,  # Try Number (Amount of attempts)
    }
    await check_limiter(*limiter)
    resp = await client.post(f'{BASE_URL}/Account/DoSolveCaptcha1', data=payload)

    if 'SolveCaptcha' in resp.url.path:
        raise PyaklassError(f'Невозможно решить капчу: "{params}".')
    return resp.text


async def fetch(client, limiter: tuple, url: str):
    """Получить текст по указанному URL адресу.

    Args:
        client (httpx.AsyncClient): Клиент
        limiter (tuple): Ограничитель и его синхронизатор
        url (str): Путь для URL: `https://www.yaklass.ru/{url}`

    Returns:
        Union[str, None]: HTML-страница или None в случае 302 (Работа еще проверяется)

    Raises:
        PossibleIPBlockError: Ошибка при возможной блокировке IP сервисом
        PyaklassError: Ошибка при выполении запроса
    """
    await check_limiter(*limiter)
    resp = await client.get(f'{BASE_URL}/{url}', timeout=None)

    if resp.status_code in (429, 403):  # Too Many Requests
        raise PossibleIPBlockError

    if resp.status_code == 302:
        # The work is being checked by a teacher. Ignore.
        return None

    if resp.status_code != 200:
        with open('error.html.bak', 'w', encoding=resp.encoding) as logf:
            logf.write(resp.text)

        raise PyaklassError(
            f'Ошибка ({resp.status_code}) при выполнении запроса {BASE_URL}/{url}.',
        )

    if 'SolveCaptcha' in resp.url.path:
        return await solve_captcha(client, limiter, resp.url.params)
    return resp.text


async def fetch_and_parse(client, limiter: tuple, url: str):
    """Получить и спарсить HTML-страницу.

    Args:
        client (httpx.AsyncClient): Клиент
        limiter (tuple): Ограничитель и его синхронизатор
        url (str): Путь для следующего URL: `https://www.yaklass.ru/{url}`

    Returns:
        Union[LexborHTMLParser, None]: Парсер или None в случае 309 (Работа еще проверяется)
    """
    text = await fetch(client, limiter, url)
    return LexborHTMLParser(text) if text else None



"""Парсинг результатов прошедшей работы."""


def _open_text_chunk(element) -> dict:
    if element.child and element.child.tag != '-text':
        text_chunk = _open_text_chunk(element.child)
    else:
        text_chunk = {'styles': [], 'text': element.html}

    if element.tag == 'strong':
        text_chunk['styles'].append('bold')
    if element.tag in {'em', 'i'}:
        text_chunk['styles'].append('italic')

    return text_chunk


def _fix_text(element) -> list:
    return [_open_text_chunk(el) for el in element.iter(include_text=True)]


def _parse_warning(element) -> dict:
    return {
        'type': 'warning',
        'title': element.css_first('.gxst-ititle > span').text(deep=False),
        'text': _fix_text(element.css_first('.gxst-ibody')),
    }


def _parse_image(element) -> dict:
    return {'type': 'image', 'src': element.css_first('.gxs-resource-image').attrs['src']}


def _parse_file_link(element) -> dict:
    # NOTE: This is not a typo.
    # For sharing similar field names between objects 'href' is named as 'src'.
    return {'type': 'file', 'src': element.css_first('.gxs-resource-file').attrs['href']}


def _parse_table(element) -> dict:
    table = element.css_first('table')

    rows = []
    current_tr = None
    for td_el in table.css('td'):
        data = [
            *(_fix_text(el) for el in td_el.iter(include_text=True) if el.tag != 'br'),
            ['\n'],
        ]

        if current_tr == td_el.parent:
            rows[-1].append(data)
        else:
            current_tr = td_el.parent
            rows.append(data)

    return {
        'type': 'table',
        'rows': rows,
    }


def parse_work_results(results: list, exercise_page):
    """Спарсить результаты прошедшей работы.

    Args:
        results (list): Результаты
        exercise_page (LexborHTMLParser): HTML-страница с заданием
    """
    title = exercise_page.css_first('.type-name').text(deep=False)

    score_el = exercise_page.css_first('.bottom-buffer-1')
    score = int(score_el.css_first('#earned-points-val').text(deep=False))
    max_score = int(score_el.text(deep=False, strip=True)[3:])

    results.append({
        'title': title,
        'body': tuple(el.html for el in exercise_page.css('#taskhtml > div > div')),
        'score': score,
        'max_score': max_score,
    })



"""Аккаунт."""
from httpx import AsyncClient

from pyaklass import parse
from pyaklass.const import BASE_URL
from pyaklass.exc import LoginError, PyaklassError
from pyaklass.testworks import get_results, get_testworks


class Account:
    """Аккаунт."""

    def __init__(self, login: str, password: str, manager=None):
        """Инициализировать аккаунт.

        Args:
            login (str): Логин
            password (str): Пароль
            manager (Union[multiprocessing.managers.SyncManager, None]): Менеджер процессов.

        `manager` может быть `None` для установки параметров вручную или через `AccountManager`.
        """
        self.__login = login
        self.__password = password

        # self.limiter[1] is a special dict for containing information about rate limits.
        if manager:
            lock = manager.Lock()
            self.limiter = (
                lock,
                manager.Value(dict, {'amount': 28, 'seconds': 10, 'current': 0}, lock),
            )
        else:
            self.limiter = (None, None)

        self.__is_logged_in = False
        self.client = AsyncClient(headers={
            'Connection': 'keep-alive',
            'sec-ch-ua': '"Google Chrome";v="108", "Chromium";v="108", "Not=A?Brand";v="24"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"Windows"',
            'Upgrade-Insecure-Requests': '1',
            'User-Agent': (
                'Mozilla/5.0 (Windows NT 10.0; Win64; x64) '
                'AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36'
            ),
            'Accept': (
                'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,'
                'image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7'
            ),
            'Sec-Fetch-Site': 'none',
            'Sec-Fetch-Mode': 'navigate',
            'Sec-Fetch-User': '?1',
            'Sec-Fetch-Dest': 'document',
            'Accept-Encoding': 'gzip, deflate',
            'Accept-Language': 'ru-RU,ru;q=0.9,en-US;q=0.8,en;q=0.7',
            'X-HTTPS': '1',
        }, follow_redirects=True)

    def get_login(self) -> str:
        """Вернуть логин.

        Returns:
            str: Логин
        """
        return self.__login

    def set_login(self, value: str):
        """Изменить логин для входа.

        Args:
            value (str): Новое значение
        """
        self.__login = value

    def set_password(self, value: str):
        """Изменить пароль для входа.

        Args:
            value (str): Новое значение
        """
        self.__password = value

    async def login(self):
        """Вход в аккаунт.

        Raises:
            LoginError: Ошибка входа в аккаунт
            ValueError: Вход в аккаунт уже был произведен
            TypeError: Пароль или логин пуст / Менеджер процессов не был указан при инициализации
        """
        if self.__is_logged_in:
            raise ValueError('Вход в аккаунт уже был произведен.')
        if not self.__login or not self.__password:
            raise TypeError('Логин или пароль пуст.')
        if not any(self.limiter):
            raise TypeError((
                'Менеджер процессов не был указан при инициализации. '
                'Для ручного создания ограничителя переопределите его значение.'
            ))

        try:
            login_page = await parse.func.fetch_and_parse(
                self.client, self.limiter, 'Account/Login',
            )
        except PyaklassError as exc:
            raise LoginError(exc.message) from exc

        crsf_token = login_page.css_first('#PostToken').attrs['value']
        payload = {
            'UserName': self.__login,
            'Password': self.__password,
            'RememberMe': 'true',
            'ReturnUrl': '',
            'AuthAction': 'lor',  # Login Or Register
            'RedirectReason': '',
            'PostToken': crsf_token,
        }
        await parse.func.check_limiter(*self.limiter)
        resp = await self.client.post(f'{BASE_URL}/Account/Login', data=payload)
        if resp.url.path == '/Account/Login':
            raise LoginError('Неправильный логин или пароль.')

        self.__password = ''  # noqa: S105
        self.__is_logged_in = True

    async def get_new_testworks(self) -> list:
        """Получить список активных работ.

        Returns:
            list: Работы
        """
        return await get_testworks(self.client, self.limiter, remove_old=True)

    async def get_all_testworks(self) -> list:
        """Получить список всех работ.

        Returns:
            list: Работы
        """
        return await get_testworks(self.client, *self.limiter)

    async def get_results(self, testwork_id: int, result_id: int) -> list:
        """Получить результаты прошедшей работы.

        Args:
            testwork_id (int): ID прошедшей работы
            result_id (int): ID результатов для работы

        Returns:
            list: Результаты
        """
        return await get_results(self.client, self.limiter, testwork_id, result_id)

    def __str__(self) -> str:
        """Вернуть удобно-читаемое представление объекта.

        Returns:
            str: Логин
        """
        return f'Аккаунт {self.__login}'



"""Менеджер для работы с несколькими аккаунтами одновременно."""
import asyncio
from functools import partial
from multiprocessing import Pool
from multiprocessing.managers import SyncManager

from httpx import AsyncClient
from selectolax.lexbor import LexborHTMLParser

from pyaklass import Account
from pyaklass.exc import LoginError
from pyaklass.parse.func import fetch
from pyaklass.testworks import get_testworks


def _w_get_new_testworks(args, limiter: tuple) -> list:
    client = AsyncClient(headers=args[0]['headers'], cookies=args[0]['cookies'])
    testwork_page = LexborHTMLParser(args[1])

    loop = asyncio.new_event_loop()
    return loop.run_until_complete(
        get_testworks(client, limiter, testwork_page, remove_old=True),
    )


def _w_get_all_testworks(args, limiter: tuple):
    client = AsyncClient(headers=args[0]['headers'], cookies=args[0]['cookies'])
    testwork_page = LexborHTMLParser(args[1])

    loop = asyncio.new_event_loop()
    return loop.run_until_complete(get_testworks(client, limiter, testwork_page))


class AccountManager:
    """Менеджер для работы с несколькими аккаунтами одновременно."""

    def __init__(self, *accounts: Account, procs: int = None):
        """Инициализировать менеджер аккаунтов.

        Args:
            accounts (tuple[Account]): Аккаунты, в которые не вошли
            procs (int): Количество процессов для параллельного парсинга
        """
        self.accounts = list(accounts)
        self.procs = procs

        # An account manager will create a global SyncManager for managing shared objects.
        self._manager = SyncManager()
        self.limiter = None

    async def prepare(self) -> list:
        """Подготовить аккаунты.

        Returns:
            list: Результат (`bool`) входа в аккаунты
        """
        self._manager.start()  # pylint: disable=consider-using-with

        lock = self._manager.Lock()
        self.limiter = (
            lock,
            self._manager.Value(dict, {'amount': 28, 'seconds': 10, 'current': 0}, lock),
        )

        async def login_or_pop(account):
            account.limiter = self.limiter

            try:
                await account.login()
                return True
            except LoginError:
                self.accounts.remove(account)
                return False

        return await asyncio.gather(*(login_or_pop(account) for account in self.accounts))

    async def _get_testworks(self, pool_func) -> list:
        with Pool(processes=self.procs) as pool:
            # NOTE: We obtain pages here instead of doing it in account.get_active_testworks(),
            # because parsing is a synchronous operation and asyncio.gather cannot perform it
            # in parallel.
            testwork_pages = await asyncio.gather(*(
                fetch(acc.client, self.limiter, '/testwork?from=menu') for acc in self.accounts
            ))
            return pool.map(
                partial(pool_func, limiter=self.limiter),
                zip(
                    [{
                        'headers': dict(account.client.headers),
                        'cookies': dict(account.client.cookies),
                    } for account in self.accounts],
                    testwork_pages,
                ),
            )

    async def get_new_testworks(self) -> list:
        """Получить список активных работ.

        Returns:
            list: Список работ для каждого аккаунта
        """
        return await self._get_testworks(_w_get_new_testworks)

    async def get_all_testworks(self) -> list:
        """Получить список всех работ.

        Returns:
            list: Список работ для каждого аккаунта
        """
        return await self._get_testworks(_w_get_all_testworks)

    def release(self):
        """Освободить ресурсы, занятые обьектами менеджера."""
        self._manager.shutdown()

    def __str__(self) -> str:
        """Вернуть удобно-читаемое представление объекта.

        Returns:
            str: Краткая информация
        """
        return f'Менеджер, работающий с {len(self.accounts)} аккаунтами'
